#ifndef _PJavaClassInfo_h
#define _PJavaClassInfo_h

#pragma interface
#include "JavaTypeSignature.h"
#include "JavaMethodSignature.h"
#include "JavaFieldSignature.h"
#include "AuxMethodMap.h"
#include "JavaAccessFlags.h"
#include "unicode_string.h"
#include <vector>
#include <functional>
#include <map>
class CCompiler;
class ostream;
class ifstream;
class CPJavaClassInfo;
class CPJavaConstraint; 
class CJavaClassFile;
class CInstantiator;

typedef map<unicode_string, CPJavaConstraint *, less<unicode_string> > 
   ParamTable;

struct TranslatedMethod {
  int fIndex;
  CJavaMethodSignature fTranslatedSignature;
  CJavaAccessFlags fAccessFlags;
};

typedef map<CJavaMethodSignature, TranslatedMethod, 
            less<CJavaMethodSignature> > MethodNameMap;

//
// Class name : CPJavaClassInfo
// Description: This class encapsulates the information bound to each 
//    parameterized class.  Ideally, we'd stick it in CJavaClassFile, but
//    since we're not integrated fully into the compilation process, we'll
//    keep it separate.
//

class CPJavaClassInfo {
public:
  // Methods
  CPJavaClassInfo(bool finished = false);
  CPJavaClassInfo(CPJavaClassInfo &other);
  CPJavaClassInfo(const unicode_string &className,
		  deque<unicode_string> *adoptParameters,
		  CJavaTypeSignature *super,
		  deque<CJavaTypeSignature>* interfaces,
		  deque<CPJavaConstraint*>* adoptConstraints,
		  bool finished = false);
  
  void SetParameters(deque<unicode_string>* adoptParameters,
		     deque<CPJavaConstraint*>* adoptConstraints);

  bool IsParameterized() const 
    { return (fParams != 0 && fParams->size() > 0); } 
  deque<unicode_string>::const_iterator ParamsBegin() const 
    { return fParams->begin(); }
  deque<unicode_string>::const_iterator ParamsEnd() const 
    { return fParams->end(); }
  const CPJavaConstraint* LookupConstraint(const unicode_string&) const;
  
  CJavaTypeSignature *GetSuper() const
    { return fSuper; }

  deque<CJavaTypeSignature>::const_iterator InterfacesBegin() const
    { assert(fInterfaces); return fInterfaces->begin(); }
  deque<CJavaTypeSignature>::const_iterator InterfacesEnd() const
    { assert(fInterfaces); return fInterfaces->end(); }

  bool ParseFromString(CJavaClassFile &classFile, const string& s);
  string WriteToString(CJavaClassFile &classFile);

  CAuxMethodMap &GetAMM() const
    { return *fAMM; }

  void RegisterConstraints();

  const MethodNameMap* GetMethods() const
    { return fMethods; }

  void Disassemble(ostream &toStream) const;

  ~CPJavaClassInfo();

protected:
  friend class CCompiler;
  friend class CInstantiator;

  string *CheckConstraints();
  bool IsFinished() const   { return fFinished; }
  bool IsProcessing() const { return fProcessing; }
  void BeginProcessing() { fProcessing = true; }
  void Finish()          { fFinished = true; fProcessing = false; }

  deque<unicode_string> *fParams; // What are this class's parameters?
  CJavaTypeSignature *fSuper;     // What is our _real_ supertype?
  deque<CJavaTypeSignature> *fInterfaces; // What are our _real_ interfaces?
  bool fHasStinit;                // Does this class have a $stinit function?
  ParamTable *fPTable;            // What are the parameters' constraints?
  CAuxMethodMap *fAMM;            // What auxmethods do we need to know about?
  MethodNameMap *fMethods;        // What are the mangled names of our
                                  //     methods? (if any.)
  bool fFinished;                 // For use by CCompiler: have we finished
                                  //     processing this one?
  bool fProcessing;               // For CCompiler: are we processing this one?
};  

#endif // _PJavaClassInfo_h

