
#ifndef _PJavaConstrint_h
#define _PJavaConstrint_h
#pragma interface

#include "JavaMethodSignature.h"
class CAuxMethodMap;

//
// Class name : CPJavaConstraint
// Description : This class encodes the information for a constraint on
//     a PJava class, as expressed in a single where clause; i.e. a
//     list of { MethodSignature, AccessFlags } where AccessFlags is either
//     "static" or nothing.
//
class CPJavaConstraint {
public:
  struct ConstraintMethod { 
    bool fStatic; 
    CJavaMethodSignature *fMethod; 
    deque<CJavaTypeSignature> *fThrows;
  };

  CPJavaConstraint();
  CPJavaConstraint(const unicode_string &param);
  CPJavaConstraint(const CPJavaConstraint &source);
  ~CPJavaConstraint();
  
  CPJavaConstraint& operator=(const CPJavaConstraint& source);

  void AddMethod(CJavaMethodSignature &method, 
		 bool isStatic, 
		 deque<CJavaTypeSignature> *adoptThrows = 0);
  void AddMethod(ConstraintMethod *method);
  void SetParameter(unicode_string &name) { fParam = name; }
  const unicode_string &GetParameter() { return fParam; }
  
  deque<ConstraintMethod*>::const_iterator MethodsBegin() const 
    { return fMethods.begin(); }
  deque<ConstraintMethod*>::const_iterator MethodsEnd() const
    { return fMethods.end(); }
  
  void Merge(const CPJavaConstraint& other);
  void Register(CAuxMethodMap& AMM);
  void Unparse(ostream &out) const;
private:
  friend class CInstantiator;
  friend class CCompiler;
  deque<ConstraintMethod*> fMethods;
  unicode_string fParam;
};


ostream& operator<<(ostream& o, const CPJavaConstraint& c);

#endif // _PJavaConstraint_h
