// Copyright (c) 1996  David Engberg  All rights reserved
// $Id: CompileContext.h,v 1.5 1998/08/20 18:28:57 nickm Exp $
#ifndef _CompileContext_h
#define _CompileContext_h
#pragma interface

#include "JavaTypeSignature.h"
#include "Bitset.h"
#include <deque>
class CCompiler;
class CJavaClassFile;
class CJavaMethodInfo;
class CSourceFile;

//
//  Class name : CCompileContext
//  Description : This is a big catch-all class that encapsulates the inherited
//    attributes used during compilation.  For example, code generation
//    occasionally needs to know what class the code is being compiled in,
//    so the compile context provides that information.  Conceptually, this
//    is information that primarily flows top-down.
//
class CCompileContext {
public:
  CCompileContext(CCompiler* compiler, CJavaClassFile* inClass,
		  CJavaMethodInfo* inMethod, 
		  CSourceFile *sourceFile,
		  unsigned long localVariablesUsed,
		  bool staticCode);
  CCompileContext(const CCompileContext& source);
  CCompileContext(const CCompileContext& source, bool staticCode);
  ~CCompileContext();
  CCompileContext& operator=(const CCompileContext& source);
  
  void Merge(const CCompileContext& other);

  CCompiler& GetCompiler() { return *fCompilerAlias; }
  const CCompiler& GetCompiler() const { return *fCompilerAlias; }

  CJavaClassFile& GetClass() { return *fClassAlias; }
  const CJavaClassFile& GetClass() const { return *fClassAlias; }

  CJavaMethodInfo& GetMethod() { return *fMethodAlias; }
  const CJavaMethodInfo& GetMethod() const { return *fMethodAlias; }
  bool HasMethod() const { return (fMethodAlias != 0); }

  bool IsVariableInitialized(unsigned long index) const;
  void InitializeVariable(unsigned long index);

  bool IsReachable() const { return fReachable; }
  void SetUnreachable() { fReachable = false; }
  void SetReachable() { fReachable = true; }

  bool Throwable(const CJavaTypeSignature& type) const;
  void PushThrowable(const CJavaTypeSignature& type);
  void PopThrowable();

  bool InStaticCode() const
    { return fInStaticCode; }

  CSourceFile* GetSource() const { return fSourceFile; }

  string GetSourceFileName() const;
private:
  CCompiler* fCompilerAlias;
  CJavaClassFile* fClassAlias;
  CJavaMethodInfo* fMethodAlias;
  CBitset fLocalVariablesInitialized;
  bool fReachable;
  bool fInStaticCode;
  deque<CJavaTypeSignature> fThrowable;
  CSourceFile *fSourceFile;
};

#endif
