// Copyright (c) 1996  David Engberg  All rights reserved
// $Id: IntermediateClass.h,v 1.11 2000/07/03 18:53:31 andru Exp $
#ifndef _IntermediateClass_h
#define _IntermediateClass_h
#pragma interface

#include "JavaClassFile.h"
#include <deque>
#include <set>
#include "unicode_string.h"
#include "JavaAccessFlags.h"
#include "parser_decls.h"
#include "Statement.h" // For SIC.
class CCompiler;
class CIntermediateFunction;
class CCompoundStatement;
class CPJavaClassInfo;
class CSourceFile;

//
//  Class name : CIntermediateClass
//  Description : This represents a Java class during the process of
//    compilation.  It's not intended as a usable end-value, but rather
//    as an intermediate phase, tightly bound to the mechanics of compilation.
//    Therefore, you can probably ignore this class unless you're playing with
//    compilation, which is why this class is 'friend CCompiler'
//    For all intents and purposes, this is basically a protected struct.
//
class CIntermediateClass {
  friend class CCompiler;
public:
  unicode_string GetName() const { return fName; }
  unicode_string GetShortName() const;
  CPJavaClassInfo * GetPInfo() const { return fPClassInfo; }
  string Disassemble() const;
  CSourceFile* GetSource()  { return fSourceFile; }
protected:
  CIntermediateClass(CSourceFile* sourceFile,
		     const string& sourceFileName,
		     const unicode_string& className,
		     CJavaAccessFlags* adoptModifiers,
		     CJavaTypeSignature* adoptExtends,
		     deque<CJavaTypeSignature>* adoptInterfaces,
		     bool deprecated,
		     unsigned long startLineNumber);
  ~CIntermediateClass();
  
  bool IsParameterized() const { return (fPClassInfo != 0) &&
	 			        (fPClassInfo->IsParameterized()); } 
  const CJavaTypeSignature &GetType() const
    { if (!fType) ((CIntermediateClass*)this)->SetType(); return *fType; }

  unsigned long GetStartLineNumber() const
    { return fStartLineNumber; }

private:
  CSourceFile *fSourceFile;
  string fSourceFileName;
  unicode_string fName;
  CJavaAccessFlags fAccessFlags;
  unicode_string* fExtends;
  deque<unicode_string> fInterfaces;
  deque<CIntermediateFunction*> fFunctions;
  unsigned short fStaticLocalVariables;
  CCompoundStatement* fStaticInitializer;
  StatementList fStaticDeclarations;
  StatementList fNonStaticDeclarations;
  CJavaClassFile* fRealClass;
  bool fDeprecated;
  unsigned long fStartLineNumber;
  unsigned long fEndLineNumber;
  
  void SetType();

  CPJavaClassInfo* fPClassInfo;
  CJavaTypeSignature* fType;
};

ostream& operator<<(ostream &o, const CIntermediateClass &sig); 

#endif
