// Copyright (c) 1996  David Engberg  All rights reserved
// $Id: IntermediateFunction.C,v 1.7 1998/07/30 16:51:30 nickm Exp $
#pragma implementation
#include "find_leaks.h"
#include "IntermediateFunction.h"
#include "Statement.h"

//
//  Method name : CIntermediateFunction
//  Description : Constructs an intermediate function out of its parts.
//
CIntermediateFunction::CIntermediateFunction(
			const CJavaMethodSignature& signature,
			const CJavaAccessFlags& modifiers,
			deque<CJavaFieldSignature>* adoptParameters,
			deque<CJavaTypeSignature>* adoptThrows,
			unsigned long startLineNumber,
			bool deprecated)
  : fSignature(signature),
    fAccessFlags(modifiers),
    fMaxLocalVariables(0),
    fBlock(0),
    fStartLineNumber(startLineNumber),
    fDeprecated(deprecated),
    fEThrows(adoptThrows)
{
  if (adoptParameters != 0) {
    fParameters = *adoptParameters;
    delete adoptParameters;
  }
  if (adoptThrows != 0) {
    for(deque<CJavaTypeSignature>::iterator i = adoptThrows->begin();
        i != adoptThrows->end(); ++i) {
      assert((*i).IsReference() && ! (*i).IsArray());
      unicode_string s;
      (*i).GetBaseClassName(s);
      fThrows.push_back(s);
    }
  }
}

//
//  Method name : ~CIntermediateFunction
//  Description : Destructor
//
CIntermediateFunction::~CIntermediateFunction()
{
  delete fBlock;
  delete fEThrows;
}

//
//  Method name : GetThrows
//  Description : This method returns a reference to a deque of all the
//     exceptions thrown by this method.
//
const deque<CJavaTypeSignature> &
CIntermediateFunction::GetThrows() const
{
  if (fEThrows == 0) {
    deque<CJavaTypeSignature> *throws = new deque<CJavaTypeSignature>;
    for (deque<unicode_string>::const_iterator i = fThrows.begin();
	 i != fThrows.end(); ++i) {
      CJavaTypeSignature t(*i);
      throws->push_back(t);
    }
    // Cast away const
    ((CIntermediateFunction*) this)->fEThrows = throws;
  }
  return *fEThrows;
}
