#ifndef _SOURCEFILE_H
#define _SOURCEFILE_H

#ifdef __GNUG__
#pragma interface
#endif
#include "JavaDirectory.h"
#include <string>
#include <deque>
#include <map>
#include <list>
#include <vector>
#include "unicode_string.h"

class CCompiler;
class CJavaClassFile;
class CSourceFile;
class CIntermediateClass;

typedef map<unicode_string, const CJavaClassFile*, less<unicode_string> >
   ImportTable;
typedef list<CJavaDirectory> ClassPathList;
typedef deque<CIntermediateClass*> IntermediateList;
typedef deque<CSourceFile*> SourceFileList;

//
//  This is basically a protected struct the elements and imports of a
//  source file.  Separating this from CCompiler allows better
//  semantics: namely, the compiler's name-lookup mechanism is not
//  affected by which files it's seen, and so can handle files in a
//  proper parse-all, write-all order.
//
class CSourceFile {
protected:
  friend class CCompiler;
  CSourceFile();
  ~CSourceFile();
  string fFileName;
  string fFilePath;  
  ClassPathList fImportedPackages;

  IntermediateList fIntermediateClasses;
  ImportTable fImportAliases;
  
  unicode_string fPackageName;
  unsigned long fPackageLineNumber;

  deque<unicode_string> fClassImports;
  vector<unsigned long> fClassImportLocations;
  deque<unicode_string> fPackageImports;
  vector<unsigned long> fPackageImportLocations;

  bool fIsPJFile;
  bool fIsValid;
  bool fResolved;
};

#endif
