/* Primitive.java
 * Base class for all PolyJ primitive type wrappers.
 */

package polyj.lang;

abstract public class Primitive {
  public Primitive() {}

  abstract public long longValue();
  abstract public double doubleValue();

  public final boolean isNumeric() {
    return ((this instanceof ByteWrap) ||
	    (this instanceof DoubleWrap) ||
	    (this instanceof FloatWrap) ||
	    (this instanceof IntWrap) ||
	    (this instanceof LongWrap) ||
	    (this instanceof ShortWrap));
  }
       
  public final boolean isFloat() {
    return ((this instanceof DoubleWrap) ||
	    (this instanceof FloatWrap));
  }

  public final boolean isInteger() {
    return ((this instanceof ByteWrap) ||
	    (this instanceof IntWrap) ||
	    (this instanceof LongWrap) ||
	    (this instanceof ShortWrap));
  }
  
  /* Checks to see whether a and b are equal.
   */
  public static boolean ee(Object a, Object b) {
    boolean aPrim = a instanceof Primitive;
    boolean bPrim = b instanceof Primitive;
    if (!aPrim && !bPrim) 
      return a == b;
    if (aPrim != bPrim)
      return false;
    Primitive aP = (Primitive) a;
    Primitive bP = (Primitive) b;
    if (aP.isFloat()) {
      return bP.isNumeric() && aP.doubleValue() == bP.doubleValue();
    } else if (bP.isFloat()) {
      return aP.isNumeric() && aP.doubleValue() == bP.doubleValue();
    } else if (aP.isNumeric()) {
      return bP.isNumeric() && aP.longValue() == bP.longValue();
    }
    // Is 'a' a boolean?
    if (a instanceof BoolWrap) {
      return (b instanceof BoolWrap) && 
	(((BoolWrap) a).value == ((BoolWrap) b).value);
    }
    if (a instanceof CharWrap) {
      return (b instanceof CharWrap) &&
	(((CharWrap) a).value == ((CharWrap) b).value);
    }
    return false;
  }
}
